'''
MIT License

Copyright 2020 Sophie Winter

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

import logging
import re

from ast import *

logger = logging.getLogger(__name__)

def remove_comments(code):
    return re.sub(r'(/\*.*?\*/|//.*?\n)', '', code, flags=re.DOTALL)

def parse_type(code):
    code = code.strip()
    if code.endswith('*'):
        return PtrType(parse_type(code[:-1]))
    elif code.startswith('const '):
        return ConstType(parse_type(code[5:]))
    if code.endswith(' const'):
        return ConstType(parse_type(code[:5]))
    elif is_std_type(code):
        return StdType(code)
    elif re.search(r'^((struct|enum)\s+)?\w+$', code):
        return CustomType(code)
    else:
        assert False, 'Unknown type `' + code + '`'

def parse_property(code):
    code = code.strip()
    bit_field = None
    bit_field_split = code.rsplit(':', 1)
    if len(bit_field_split) > 1:
        code = bit_field_split[0].strip()
        bit_field = int(bit_field_split[1].strip())
    fp = re.search(r'^(.*)\(\s*(\*\s*)+(\w*)\s*\)\s*\((.*)\)$', code)
    if fp:
        assert bit_field is None
        ret = parse_type(fp.group(1))
        pointer_count = fp.group(2).count('*') - 1
        assert pointer_count == 0, (
            'Function pointers with multiple levels of indirection not supported '
            '(indirection count: ' + str(pointer_count) + ' for ' + code + ')')
        name = fp.group(3)
        if name == '':
            name = None
        args = []
        for i in fp.group(4).split(','):
            args.append(parse_property(i))
        c_type = FpType(ret, args)
        return PropertyNode(c_type, name, None)
    array = re.search(r'^(.*)\[(.*)\]$', code)
    if array:
        prop = parse_property(array.group(1))
        prop.c_type = ArrayType(prop.c_type, array.group(2).strip())
        return prop
    normal = re.search(r'^((\w+\s*[^\w\,])+(\*\s*)*)([\w\s\,]*)$', code)
    if normal:
        c_type = parse_type(normal.group(1))
        names = normal.group(4).split(',')
        names = [name.strip() for name in names if name.strip()]
        if len(names) > 1:
            nodes = []
            for name in names:
                nodes.append(PropertyNode(c_type, name, None))
                nodes[-1].statement = True
            return ListNode(nodes)
        elif len(names) == 1:
            return PropertyNode(c_type, names[0], bit_field)
        else:
            return PropertyNode(c_type, None, bit_field)
    assert False, 'Could not parse property `' + code + '`'

def parse_token_list(tokens, i):
    nodes = []
    while i < len(tokens):
        token = tokens[i].strip()
        if token == '' or token == ';':
            pass
        elif token == 'struct' and tokens[i + 1].strip() == '{':
            i, node = parse_token_list(tokens, i + 2)
            i += 1
            assert i < len(tokens)
            name = tokens[i].strip()
            assert re.search(r'^\w+$', name)
            struct = SubStructNode(name, node)
            nodes.append(struct)
        elif token == '}':
            break
        else:
            node = parse_property(token)
            if isinstance(node, PropertyNode):
                node.statement = True
            nodes.append(node)
        i += 1
    return i, ListNode(nodes)

def parse_ast(code):
    code = remove_comments(code)
    code = code.replace('\n', ' ')
    code = code.replace('\t', ' ')
    tokens = re.split(r'([;\{\}])', code)
    i, node = parse_token_list(tokens, 0)
    assert i == len(tokens), code
    return node



'''
MIT License

Copyright 2020 Sophie Winter

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

import logging
import functools
import re

from config import MIN_SUPPORTED_GTK, MAX_SUPPORTED_GTK

logger = logging.getLogger(__name__)

COMBO_FACTOR = 1000

@functools.total_ordering # implement all comparisons with just eq and lt
class Version:
    def __init__(self, git_name, minor, micro, released):
        self.git_name = git_name
        self.minor = minor
        self.micro = micro
        self.released = released

    def __eq__(self, other):
        return self.minor == other.minor and self.micro == other.micro

    def __lt__(self, other):
        if self.minor == other.minor:
            return self.micro < other.micro
        else:
            return self.minor < other.minor

    def get_checkout_name(self):
        return self.git_name

    def is_supported(self):
        '''Returns if the version is one we support'''
        return (
            self >= min_supported_version and
            self <= max_supported_version and
            self != bad_release_3_24_19)

    def is_released(self):
        return self.released

    def get_combo(self):
        return self.minor * 1000 + self.micro

    def __str__(self):
        result = 'v3.' + str(self.minor) + '.' + str(self.micro)
        if not self.is_released():
            result += ' (unreleased)'
        return result

    def c_id(self):
        '''a string suitable for a C identifier'''
        return 'v3_' + str(self.minor) + '_' + str(self.micro)

def parse_tag(tag):
    match = re.search(r'^3\.(\d+)\.(\d+)$', tag)
    if match:
        minor = int(match.group(1))
        micro = int(match.group(2))
        return Version(tag, minor, micro, True)
    else:
        return None

def parse_branch(branch):
    match = re.search(r'^gtk\-3\-(\d+)$', branch)
    if match:
        minor = int(match.group(1))
        return minor
    else:
        return None

min_supported_version = parse_tag(MIN_SUPPORTED_GTK)
max_supported_version = parse_tag(MAX_SUPPORTED_GTK)
bad_release_3_24_19 = parse_tag('3.24.19') # this is not a good release

def parse_tags_and_branches(tags, branches):
    result = []
    for tag in tags:
        version = parse_tag(tag)
        if version and version.is_supported():
            result.append(version)
    highest_micro = {}
    for v in result:
        highest_micro[v.minor] = max(v.micro, highest_micro.get(v.minor, -1))
    for branch in branches:
        minor = parse_branch(branch)
        if minor is not None:
            micro = highest_micro.get(minor, -1) + 1
            version = Version(branch, minor, micro, False)
            if version.is_supported():
                result.append(version)
    result.sort()
    logger.info('Found ' + str(len(result)) + ' supported versions')
    return result



'''
MIT License

Copyright 2020 Sophie Winter

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

import logging
import re

from version import Version
import code

logger = logging.getLogger(__name__)

def remove_comments(code):
    return re.sub(r'(/\*.*?\*/|//.*?\n)', '', code, flags=re.DOTALL)

# These can be combined in any order, we do not check that it is valid
std_types = set([
    'void',
    'int',
    'char',
    'signed',
    'unsigned',
    'short',
    'long',
    'float',
    'double',
    'uint32_t',
    'gint',
    'guint',
    'gchar',
    'gboolean',
    'gpointer',
])

def is_std_type(name):
    for i in name.split():
        if not i in std_types:
            return False
    return True

class CType:
    def __str__(self):
        return self.str_left(False) + self.str_right(False)

    def str_right(self, resolved):
        return ''

    def __eq__(self, other):
        return str(self) == str(other)

class StdType(CType):
    def __init__(self, name):
        assert isinstance(name, str)
        assert is_std_type(name), name
        self.name = name

    def resolve(self, ctx):
        pass

    def str_left(self, resolved):
        return self.name

class CustomType(CType):
    def __init__(self, name):
        assert isinstance(name, str)
        self.explicit_struct = False
        if name.startswith('struct'):
            name = name[6:].strip()
            self.explicit_struct = True
        self.explicit_enum = False
        if name.startswith('enum'):
            name = name[4:].strip()
            self.explicit_enum = True
        self.name = name
        self.version = None
        self.resolved_struct = None

    def resolve(self, ctx):
        if self.explicit_enum:
            struct = None
        elif self.explicit_struct:
            struct = ctx.project.lookup_struct_name(self.name)
        else:
            struct = ctx.project.lookup_typedef(self.name)

        if struct:
            self.resolved_struct = struct
            self.version = ctx.version

    def str_left(self, resolved):
        if self.resolved_struct and resolved:
            v = self.resolved_struct.lookup_version(self.version)
            return 'struct ' + v.versioned_struct_name()
        else:
            result = ''
            if self.explicit_struct:
                result += 'struct '
            if self.explicit_enum:
                result += 'enum '
            result += self.name
            return result

class PtrType(CType):
    def __init__(self, inner):
        assert isinstance(inner, CType)
        self.inner = inner

    def resolve(self, ctx):
        self.inner.resolve(ctx)

    def str_left(self, resolved):
        result = self.inner.str_left(False)
        if not isinstance(self.inner, PtrType):
            result += ' '
        result += '*'
        return result

    def str_right(self, resolved):
        return self.inner.str_right(False)

class ConstType(CType):
    def __init__(self, inner):
        assert isinstance(inner, CType)
        self.inner = inner

    def resolve(self, ctx):
        self.inner.resolve(ctx)

    def str_left(self, resolved):
        return 'const ' + self.inner.str_left(resolved)

    def str_right(self, resolved):
        return self.inner.str_right(resolved)

# Note the first (and usually only) layer of "pointer" is consumed by this type
# Do not wrap in a PtrType in most cases
class FpType(CType):
    def __init__(self, return_type, arg_list):
        assert isinstance(return_type, CType)
        assert isinstance(arg_list, list)
        for i in arg_list:
            assert isinstance(i, PropertyNode), str(i) + ' is not a property node'
        self.return_type = return_type
        self.arg_list = arg_list

    def resolve(self, ctx):
        self.return_type.resolve(ctx)
        for arg in self.arg_list:
            arg.resolve(ctx)

    def str_left(self, resolved):
        return str(self.return_type) + ' (*'

    def str_right(self, resolved):
        return ') (' + str(', '.join([str(t) for t in self.arg_list])) + ')'

class ArrayType(CType):
    def __init__(self, inner, size):
        assert isinstance(inner, CType)
        assert isinstance(size, str)
        self.inner = inner
        self.size = size

    def resolve(self, ctx):
        self.inner.resolve(ctx)

    def str_left(self, resolved):
        if resolved:
            return self.inner.str_left(True)
        else:
            return self.inner.str_left(False) + '*'

    def str_right(self, resolved):
        if resolved:
            return '[' + self.size + ']' + self.inner.str_right(True)
        else:
            return self.inner.str_right(False)

class AstNode:
    def __eq__(self, other):
        return str(self) == str(other)

class PropertyNode(AstNode):
    def __init__(self, c_type, name, bit_field):
        assert isinstance(c_type, CType)
        assert name is None or isinstance(name, str)
        assert bit_field is None or isinstance(bit_field, int)
        self.c_type = c_type
        self.name = name
        self.statement = False
        self.bit_field = bit_field

    def resolve(self, ctx):
        self.c_type.resolve(ctx)

    def get_property_list(self, prefix):
        return [(self.c_type, prefix + self.name)]

    def __str__(self):
        result = self.c_type.str_left(True)
        if self.name:
            if not isinstance(self.c_type, PtrType):
                result += ' '
            result += self.name
        result += self.c_type.str_right(True)
        if self.bit_field:
            result += ' : ' + str(self.bit_field)
        if self.statement:
            result += ';\n'
        return result

class ListNode(AstNode):
    def __init__(self, nodes):
        assert isinstance(nodes, list)
        for node in nodes:
            assert isinstance(node, AstNode)
        self.nodes = nodes

    def resolve(self, ctx):
        for node in self.nodes:
            node.resolve(ctx)

    def get_property_list(self, prefix):
        result = []
        for node in self.nodes:
            result += node.get_property_list(prefix)
        return result

    def __str__(self):
        return ''.join(str(node) for node in self.nodes)

class SubStructNode(AstNode):
    def __init__(self, name, content):
        assert isinstance(name, str)
        assert isinstance(content, AstNode)
        self.name = name
        self.content = content

    def resolve(self, ctx):
        self.content.resolve(ctx)

    def get_property_list(self, prefix):
        return self.content.get_property_list(prefix + self.name + '.')

    def __str__(self):
        return (
            'struct {' +
            ('\n' + code.INDENT).join([''] + str(self.content).splitlines()) +
            '\n} ' + self.name + ';\n')



'''
MIT License

Copyright 2020 Sophie Winter

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

import logging
from datetime import datetime
import os
from os import path
import re
import mmap
from collections import OrderedDict
import tempfile
import subprocess

import parse
from ast import *
from version import COMBO_FACTOR, Version

logger = logging.getLogger(__name__)

INDENT = '  ' # used for generating code
PROJECT_NAME = 'gtk-layer-shell'

LGPL3_HEADER = '''
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
'''

CODE_EXTENSIONS = set(['.h', '.hpp', '.c', '.cpp', '.cc'])

def typdef_to_struct_name(typedef):
    return '_' + typedef

def get_all_source_files(search_dir):
    result = []
    for item in os.listdir(search_dir):
        p = path.join(search_dir, item)
        if path.isdir(p):
            result += get_all_source_files(p)
        elif path.splitext(p)[1] in CODE_EXTENSIONS:
            result.append(p)
    return result

def remove_headers_from_dir(header_dir):
    logger.info('Clearing header files out of ' + header_dir)
    for item in os.listdir(header_dir):
        p = path.join(header_dir, item)
        if item != 'common.h' and path.splitext(p)[1] == '.h':
            os.remove(p)

def struct_regex_string(struct_name):
    return r'struct\s+' + struct_name + r'\s*\{'

def extract_body(file_content, struct_name):
    halves = re.split(struct_regex_string(struct_name), file_content)
    if len(halves) < 2:
        raise RuntimeError(struct_name + ' not declared in ' + code_path + ' (even though it was detected)')
    if len(halves) > 2:
        raise RuntimeError(struct_name + ' declared multiple times in ' + code_path)
    body = ''
    brace_level = 1
    for c in halves[1]:
        if c == '{':
            brace_level += 1
        elif c == '}':
            brace_level -= 1
        if brace_level <= 0:
            break
        body += c
    return body

def camel_case_to_words(name):
    result = []
    word = ''
    for l in name:
        if l.isupper() and word:
            result.append(word)
            word = ''
        word += l.lower()
    if word:
        result.append(word)
    return result

def c_function(return_type, name, arg_list, body):
    result = ''
    result += return_type.str_left(False)
    result += ' ' + name + '('
    arg_strs = [c_type.str_left(False) + ' ' + name + c_type.str_right(False) for c_type, name in arg_list]
    result += ', '.join(arg_strs)
    result += ')' + return_type.str_right(False) + ' {\n'
    for line in body.strip().splitlines():
        result += INDENT + line + '\n'
    result += '}\n'
    return result

def write_tmp_file(text):
    f = tempfile.NamedTemporaryFile(mode='w', delete=True)
    f.write(text)
    f.flush()
    return f

def diff_between(old, new):
    old_f = write_tmp_file(old)
    new_f = write_tmp_file(new)
    result = subprocess.run(['diff', old_f.name, new_f.name], capture_output=True, text=True)
    code = ''
    for line in result.stdout.splitlines():
        if line.startswith('<'):
            code += '// -' + line[1:] + '\n'
        elif line.startswith('>'):
            code += '// +' + line[1:] + '\n'
    return code

class ResolveContext:
    def __init__(self, project, struct, version):
        self.project = project
        self.struct = struct
        self.version = version

class StructVersion:
    def __init__(self, code_path, project, struct, version):
        with open(code_path, 'r') as f:
            source_code = f.read()
        self.code_path = code_path
        self.first_version = version
        self.last_version = version
        self.struct_name = struct.struct_name
        self.copyright_lines = set(re.findall(r'[Cc]opyright .*(?=\n)', source_code))
        self.body = extract_body(source_code, self.struct_name)
        self.ast = parse.parse_ast(self.body)
        self.ast.resolve(ResolveContext(project, struct, version))

    def get_code_path(self):
        return self.code_path

    def get_property_list(self):
        return self.ast.get_property_list('')

    def version_range_str(self):
        if self.first_version == self.last_version:
            return str(self.first_version)
        else:
            return str(self.first_version) + ' - ' + str(self.last_version)

    def emit_definition(self, generated):
        result = ''
        result += '// Valid for GTK ' + self.version_range_str() + '\n'
        result += 'struct ' + str(self) + '\n{'
        if generated:
            result += (
                '\n' + INDENT +
                ('\n' + INDENT).join(str(self.ast).splitlines()) +
                '\n')
        else:
            result += self.body
        result += '};\n'
        return result

    def is_valid_for(self, version):
        return version >= self.first_version and version <= self.last_version

    def versioned_struct_name(self):
        return self.struct_name + '_' + self.first_version.c_id()

    def __str__(self):
        return self.versioned_struct_name()

    def __eq__(self, other):
        assert isinstance(other, StructVersion)
        return self.ast == other.ast

class Property:
    def __init__(self, struct, c_type, name, version_ids):
        assert isinstance(struct, Struct)
        assert isinstance(c_type, CType)
        assert isinstance(name, str)
        assert isinstance(version_ids, list)
        for i in version_ids:
            assert isinstance(i, bool) # If this property is supported for each version ID
        self.struct = struct
        self.c_type = c_type
        self.name = name
        self.version_ids = version_ids

    def all_versions_supported(self):
        for i in self.version_ids:
            if not i:
                return False
        return True

    def get_id_name(self):
        return self.name.replace('.', '_')

    def get_fn_name(self, action, suffix):
        result = ''
        result += '_'.join(camel_case_to_words(self.struct.typedef))
        result += '_priv_'
        if action:
            result += action + '_'
        result += self.get_id_name()
        if suffix:
            result += '_' + suffix
        return result

    def emit_version_id_switch(self, on_supported, on_unsupported):
        result = ''
        result += 'switch (' + self.struct.get_version_id_fn_name() + '()) {\n'
        for i, supported in enumerate(self.version_ids):
            result += INDENT + 'case ' + str(i) + ': '
            type_name = 'struct ' + self.struct.versions[i].versioned_struct_name()
            if supported:
                result += on_supported(type_name)
            else:
                result += on_unsupported(type_name)
            result += '\n'
        result += INDENT + 'default: g_error("Invalid version ID"); g_abort();\n'
        result += '}'
        return result

    def emit_ptr_getter(self):
        ret_type = PtrType(self.c_type)
        suffix = 'ptr' if self.all_versions_supported() else 'ptr_or_null'
        fn_name = self.get_fn_name('get', suffix)
        arg_list = [(self.struct.get_ptr_type(), 'self')]
        body = self.emit_version_id_switch(
            lambda type_name: 'return (' + str(ret_type) + ')&((' + type_name + '*)self)->' + self.name + ';',
            lambda type_name: 'return NULL;')
        return c_function(ret_type, fn_name, arg_list, body)

    def emit_is_supported(self):
        ret_type = StdType('gboolean')
        fn_name = self.get_fn_name('get', 'supported')
        body = self.emit_version_id_switch(
            lambda type_name: 'return TRUE;',
            lambda type_name: 'return FALSE;')
        return c_function(ret_type, fn_name, [], body)

    def emit_not_supported_error(self, type_name):
        msg = '"' + self.struct.typedef + '::' + self.name + ' not supported on this GTK"'
        return 'g_error(' + msg + '); g_abort();'

    def emit_getter(self):
        ret_type = self.c_type
        suffix = '' if self.all_versions_supported() else 'or_abort'
        fn_name = self.get_fn_name('get', suffix)
        arg_list = [(self.struct.get_ptr_type(), 'self')]
        body = self.emit_version_id_switch(
            lambda type_name: 'return ((' + type_name + '*)self)->' + self.name + ';',
            self.emit_not_supported_error)
        return c_function(ret_type, fn_name, arg_list, body)

    def emit_setter(self):
        ret_type = StdType('void')
        suffix = '' if self.all_versions_supported() else 'or_abort'
        fn_name = self.get_fn_name('set', suffix)
        arg_list = [(self.struct.get_ptr_type(), 'self'), (self.c_type, self.get_id_name())]
        body = self.emit_version_id_switch(
            lambda type_name: '((' + type_name + '*)self)->' + self.name + ' = ' + self.get_id_name() + '; break;',
            self.emit_not_supported_error)
        return c_function(ret_type, fn_name, arg_list, body)

    def emit_functions(self):
        result = ''
        result += '// ' + self.struct.typedef + '::' + self.name + '\n\n'
        if not self.all_versions_supported():
            result += self.emit_is_supported()
            result += '\n'
        if isinstance(self.c_type, CustomType) or isinstance(self.c_type, ArrayType):
            result += self.emit_ptr_getter()
        else:
            result += self.emit_getter()
            result += '\n'
            result += self.emit_setter()
        return result

class Struct:
    def __init__(self, typedef):
        self.typedef = typedef
        self.struct_name = typdef_to_struct_name(typedef)
        self.versions = []
        self.supported_versions = []
        self.copyright_lines = set()
        self.search_regex = re.compile(bytes(struct_regex_string(self.struct_name), 'utf-8'))

    def lookup_version(self, version):
        for i in self.versions:
            if i.is_valid_for(version):
                return i

    def get_code_path(self):
        if self.versions:
            return self.versions[-1].get_code_path()
        else:
            return None

    def get_ptr_type(self):
        return PtrType(CustomType(self.typedef))

    def get_version_id_fn_name(self):
        return '_'.join(camel_case_to_words(self.typedef)) + '_priv_get_version_id'

    def header_name(self):
        return '_'.join(camel_case_to_words(self.typedef)) + '_priv.h'

    def macro_name(self):
        return self.header_name().replace('.', '_').upper()

    def add_version(self, new):
        self.versions.append(new)
        self.supported_versions.append(new.first_version)
        self.copyright_lines = self.copyright_lines.union(new.copyright_lines)

    # Returns if any changes were made
    def simplify(self):
        new_versions = []
        dropped_versions = 0
        for v in self.versions:
            if new_versions and new_versions[-1] == v:
                dropped_versions += 1
                new_versions[-1].last_version = v.last_version
            else:
                new_versions.append(v)
        if dropped_versions > 0:
            logger.info('Found ' + str(dropped_versions) + ' unnecessary versions of ' + self.typedef);
        self.versions = new_versions
        return dropped_versions > 0

    def setup_properties(self):
        prop_dict = OrderedDict()
        for i, v in enumerate(self.versions):
            for c_type, name in v.get_property_list():
                if name in prop_dict and not (prop_dict[name][0] == c_type):
                    logger.warning(
                        'Property ' + name +
                        ' changes type from ' + str(prop_dict[name][0]) +
                        ' to ' + str(c_type) +
                        ', ignoring old type')
                    del prop_dict[name]
                if name in prop_dict:
                    assert prop_dict[name][0] == c_type
                    prop_dict[name][1].add(i)
                else:
                    prop_dict[name] = (c_type, set([i]))
        self.properties = []
        for name, (c_type, supported_ids) in prop_dict.items():
            versions_ids = [bool(i in supported_ids) for i in range(len(self.versions))]
            self.properties.append(Property(self, c_type, name, versions_ids))

    def emit_get_version_id_fn(self):
        return_type = StdType('int')
        fn_name = self.get_version_id_fn_name()
        body = ''
        body += 'static int version_id = -1;\n'
        body += '\n'
        body += 'if (version_id == -1) {\n'
        body += INDENT + 'gtk_priv_assert_gtk_version_valid();'
        body += '\n'
        body += INDENT + 'int combo = gtk_get_minor_version() * 1000 + gtk_get_micro_version();\n'
        body += '\n'
        body += INDENT + 'switch (combo) {\n'
        for v in self.supported_versions:
            if v.is_released():
                body += INDENT * 2 + 'case ' + str(v.get_combo()) + ':\n'
        body += INDENT * 3 + 'break;\n'
        body += '\n'
        body += INDENT * 2 + 'default:\n'
        body += INDENT * 3 + 'gtk_priv_warn_gtk_version_may_be_unsupported();\n'
        body += INDENT + '}\n'
        body += '\n'
        body += INDENT
        for i in range(len(self.versions) - 1, -1, -1):
            if i > 0:
                body += 'if (combo >= ' + str(self.versions[i].first_version.get_combo()) + ') '
            body += '{\n'
            body += INDENT * 2 + 'version_id = ' + str(i) + ';\n'
            body += INDENT + '}'
            if i > 0:
                body += ' else '
        body += '\n'
        body += '}\n'
        body += '\n'
        body += 'return version_id;\n'
        return c_function(return_type, fn_name, [], body)

    def emit_header(self, generated):
        result = ''
        result += '/* AUTOGENERATED, DO NOT EDIT DIRECTLY\n'
        result += ' * See gtk-priv/README.md for more information\n'
        result += ' *\n'
        result += ' * This file is part of ' + PROJECT_NAME + '\n'
        result += ' *\n'
        me = path.relpath(__file__, path.dirname(path.dirname(path.dirname(__file__))))
        my_copyright_line = 'Copyright © ' + str(datetime.now().year) + ' ' + me
        for line in list(self.copyright_lines) + [my_copyright_line]:
            result += ' * ' + line + '\n'
        result += ' *'
        result += '\n * '.join(LGPL3_HEADER.splitlines())
        result += '\n */\n'
        result += '\n'
        result += '#ifndef ' + self.macro_name() + '\n'
        result += '#define ' + self.macro_name() + '\n'
        result += '\n'
        result += '#include "common.h"\n'
        result += '\n'
        result += 'typedef struct ' + self.struct_name + ' ' + self.typedef + ';\n'
        result += '\n'
        prev_definition = None
        for i, struct in enumerate(self.versions):
            result += '// Version ID ' + str(i) + '\n'
            definition = struct.emit_definition(generated)
            if prev_definition:
                result += '// Diff from previous version:\n'
                result += '\n'.join(diff_between(prev_definition, definition).splitlines()[4:]) + '\n'
            prev_definition = definition
            result += definition
            result += '\n'
        result += '// For internal use only\n'
        result += self.emit_get_version_id_fn()
        result += '\n'
        for p in self.properties:
            result += p.emit_functions()
            result += '\n'
        result += '#endif // ' + self.macro_name() + '\n'
        return result

def file_contains_byte_regex(source_file, regex):
    try:
        with open(source_file) as f:
            with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as s:
                return bool(regex.search(s))
    except (ValueError, FileNotFoundError):
        return False

class Project:
    def __init__(self, repo_dir, typedef_names):
        self.repo_dir = repo_dir
        structs = [Struct(typedef) for typedef in typedef_names]
        self.typedefs = {struct.typedef: struct for struct in structs}
        self.struct_names = {struct.struct_name: struct for struct in structs}

    def lookup_struct_name(self, struct_name):
        return self.struct_names.get(struct_name)

    def lookup_typedef(self, typedef):
        return self.typedefs.get(typedef)

    def update(self, version):
        source_files = None
        for typedef, struct in self.typedefs.items():
            code_path = struct.get_code_path()
            search_regex = struct.search_regex
            if not code_path or not file_contains_byte_regex(code_path, search_regex):
                if source_files is None:
                    source_files = get_all_source_files(self.repo_dir)
                files = set()
                for source_file in source_files:
                    if file_contains_byte_regex(source_file, search_regex):
                        files.add(source_file)
                if not files:
                    raise RuntimeError('Could not find ' + typedef + ' in ' + str(version))
                if len(files) > 1:
                    raise RuntimeError(typedef + ' implemented multiple places: ' + str(files))
                code_path = list(files)[0]
            struct_version = StructVersion(code_path, self, struct, version)
            struct.add_version(struct_version)

    def simplify(self):
        i = 1;
        while True:
            logger.info('Detecting identical versions, round ' + str(i))
            i += 1
            made_change = False
            for _, struct in self.typedefs.items():
                if struct.simplify():
                    made_change = True
            if not made_change:
                break
        for _, struct in self.typedefs.items():
            struct.setup_properties()

    def write(self, output_dir):
        remove_headers_from_dir(output_dir)
        for typedef, struct in self.typedefs.items():
            output_path = path.join(output_dir, struct.header_name())
            logger.info('Writing ' + str(len(struct.versions)) + ' versions of ' + typedef + ' to ' + output_path)
            with open(output_path, "w") as f:
                f.write(struct.emit_header(True))



#!/usr/bin/python3
'''
This entire file is licensed under MIT.

Copyright 2020 Sophie Winter

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

# This script checks that every source file in the project has an appropriate license

import logging
import re
import os
from os import path
import subprocess
import re

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.WARNING)

toplevel_dirs = ['include', 'src', 'gtk-priv', 'examples', 'test']
ignore_patterns_file = 'test/license-ignore.txt'

MIT_EXAMPLE = '''
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

LGPL3_EXAMPLE = '''
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
'''

def canonify_str(s):
    return re.sub(r'[/#\'\n\s\*]', '', s)

def get_project_root():
    return path.dirname(path.dirname(path.realpath(__file__)))

ignore_patterns = None
def get_ignore_patterns():
    global ignore_patterns
    if ignore_patterns is not None:
        return ignore_patterns
    ignore_patterns = []
    for f in ['.gitignore', '.git/info/exclude', ignore_patterns_file]:
        p = get_project_root() + '/' + f
        if path.isfile(p):
            logger.info('Excluding paths in ' + p + ' from the license check')
            for raw_line in open(p, 'r').read().splitlines():
                line = re.sub(r'([^#]*)(.*)', r'\1', raw_line).strip()
                if line:
                    ignore_patterns.append(line)
        else:
            logger.warning(p + ' not found, it will not be used to exclude paths from the license check')
    return ignore_patterns

def path_matches(base_path, original):
    pattern = re.escape(original)
    pattern = re.sub(r'\\\*', r'.*', pattern)
    pattern = re.sub(r'\\?/', r'(^|$|/)', pattern)
    try:
        return bool(re.match(pattern, base_path))
    except Exception as e:
        raise RuntimeError('Failed to match pattern ' + pattern + ' (original: ' + original + ') against ' + base_path + ': ' + str(e))

def get_files(prefix, search_path):
    full_path = path.join(prefix, search_path);
    assert path.exists(full_path), full_path + ' does not exist'
    if path.exists(path.join(full_path, 'build.ninja')):
        logger.info(search_path + ' ignored because it is a build directory')
        return []
    for pattern in get_ignore_patterns():
        if path_matches(search_path, pattern):
            logger.info(search_path + ' ignored because it matches ' + pattern)
            return []
    if path.isfile(full_path):
        logger.info('Found ' + search_path)
        return [search_path]
    elif path.isdir(full_path):
        logger.info('Scanning ' + search_path)
        result = []
        for item in os.listdir(full_path):
            result += get_files(prefix, path.join(search_path, item))
        return result
    else:
        return []

def get_important_files():
    result = []
    for toplevel_dir in toplevel_dirs:
        logger.info('Scanning toplevel directory ' + toplevel_dir)
        result += get_files(get_project_root(), toplevel_dir)
    return result

def print_list(name, files):
    if files:
        print('The following files are licensed under ' + name + ':')
        for f in files:
            print('  ' + f)
    else:
        print('No files are licensed under ' + name)
    print()

def load_file(p):
    try:
        with open(p, 'r') as f:
            contents = f.read()
            return canonify_str(contents)
    except Exception:
        rel = path.relpath(p, get_project_root())
        raise RuntimeError(
            'Failed to read ' + rel +
            '. If this file should have been ignored, add it to ' + ignore_patterns_file)

def main():
    logger.info('Project root: ' + get_project_root())
    logger.info('Ignore paths: \n  ' + '\n  '.join(get_ignore_patterns()))
    all_files = get_important_files()
    logger.info('Found ' + str(len(all_files)) + ' files')
    assert len(all_files) > 10, 'There are ' + str(len(all_files)) + ' files (which is not as many as there should be)'
    mit_files = []
    lgpl3_files = []
    none_files = []
    multiples_files = []
    mit_example = canonify_str(MIT_EXAMPLE)
    lgpl3_example = canonify_str(LGPL3_EXAMPLE)
    for p in all_files:
        contents = load_file(path.join(get_project_root(), p))
        found = 0
        if mit_example in contents:
            mit_files.append(p)
            found += 1
        if lgpl3_example in contents:
            lgpl3_files.append(p)
            found += 1
        if found > 1:
            multiples_files.append(p)
        elif found < 1:
            none_files.append(p)
    print()
    print_list('MIT', mit_files)
    print_list('LGPLv3', lgpl3_files)
    print_list('no license', none_files)
    print_list('multiple licenses', multiples_files)
    if none_files or multiples_files:
        print('If some files should be excluded from the license check, add them to ' + ignore_patterns_file)
        print('Failed license check')
        exit(1)
    else:
        print('Passed license check')
        exit(0)

if __name__ == '__main__':
    main()



#!/usr/bin/python3
'''
This entire file is licensed under MIT.

Copyright 2020 Sophie Winter

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

# This script runs an integration test. See test/README.md for details
usage = 'Usage: python3 run-test <test-binary>'

import os
from os import path
import sys
import shutil
import time
import subprocess
import threading
from typing import List, Dict, Optional, Any

# All callables (generally lambdas) appended to this list will be called at the end of the program
cleanup_funcs = []

class TestError(RuntimeError):
    pass

def get_xdg_runtime_dir() -> str:
    '''
    Creates a directory to use as the XDG_RUNTIME_DIR.
    It bases the result on the current PID because each test running in parallel needs a unique directory.
    '''
    tmp_runtime_dir = '/tmp/layer-shell-test-runtime-dir-' + str(os.getpid())
    if (path.exists(tmp_runtime_dir)):
        # We should wipe the dir on cleanup, but things can go wrong
        wipe_xdg_runtime_dir(tmp_runtime_dir)
    os.mkdir(tmp_runtime_dir)
    cleanup_funcs.append(lambda: wipe_xdg_runtime_dir(tmp_runtime_dir))
    return tmp_runtime_dir

def wipe_xdg_runtime_dir(p: str):
    assert p.startswith('/tmp'), 'Sanity check'
    assert 'layer-shell-test-runtime-dir' in p, 'Sanity check'
    shutil.rmtree(p)

def wait_until_appears(p: str):
    '''Waits for something to appear at the given path'''
    sleep_time = 0.01
    timeout = 5.0
    for i in range(int(timeout / sleep_time)):
        if path.exists(p):
            return
        time.sleep(0.01)
    raise TestError(p + ' did not appear in ' + str(timeout) + ' seconds')

def format_stream(name: str, stream: str) -> str:
    '''
    After collecting a programs output stream into a string, this function formats it for easy reading.
    Specifically, it gives it a colored border and a name.
    '''
    l_pad = 28 - len(name) // 2
    r_pad = l_pad
    if len(name) % 2 == 1:
        r_pad -= 1
    l_pad = max(l_pad, 1)
    r_pad = max(r_pad, 1)
    header = '─' * l_pad + '┤ ' + name + ' ├' + '─' * r_pad + '┈'
    divider = '\n│'
    body = divider.join(' ' + line for line in stream.strip().splitlines())
    footer = '─' * 60 + '┈'
    return '╭' + header + divider + body + '\n╰' + footer

def format_process_report(name: str, returncode: int, stdout: str, stderr: str) -> str:
    '''After running a program, this function is used to format it's output for easy reading'''
    result = format_stream(name + ' stderr', stderr) + '\n\n'
    if stdout:
        result += format_stream(name + ' stdout', stdout) + '\n\n'
    else:
        result += 'stdout empty, '
    result += 'exit code: ' + str(returncode)
    return result

class Pipe:
    '''
    The normal Python subprocess.PIPE freezes the subprocess after it fills up a finite output buffer.
    This class solves that by opening a pipe and reading from it from another thread whil the subprocess runs.
    This increases the output limit to the system's available memory

    See: https://github.com/wmww/gtk-layer-shell/issues/91#issuecomment-719082062
    See: https://thraxil.org/users/anders/posts/2008/03/13/Subprocess-Hanging-PIPE-is-your-enemy/
    '''
    def __init__(self, name: str):
        '''name is only for debugging'''
        readable, writable = os.pipe()
        self.fd = writable
        self.data: Any = bytes()
        self.result: Optional[str] = None
        # Read the data coming out of the pipe on a background thread
        # This keeps the buffer from filling up and blocking
        self.reader_thread = threading.Thread(name=name, target=self.read, args=(readable,))
        self.reader_thread.start()
        cleanup_funcs.append(lambda: self.close())

    def read(self, readable: int):
        '''Reads from the given fd until the other side closes'''
        while True:
            data = os.read(readable, 1000)
            if not data:
                # We've reached the end of the data
                os.close(readable)
                return
            self.data += data

    def close(self):
        '''Closes the fd and stops the reader'''
        if self.reader_thread.is_alive():
            os.close(self.fd)
            self.reader_thread.join(timeout=1)
            assert not self.reader_thread.is_alive(), 'Failed to join pipe reader thread'

    def collect_str(self) -> str:
        '''Closes the pipe if needed and returns the read data decoded as UTF-8'''
        if self.result is None:
            self.close()
            self.result = self.data.decode('utf-8')
            self.data = None
        return self.result

class Program:
    '''A program to run as a subprocess'''
    def __init__(self, name: str, args: List[str], env: Dict[str, str]):
        self.name = name
        self.stdout = Pipe(name + ' stdout')
        self.stderr = Pipe(name + ' stderr')
        self.subprocess = subprocess.Popen(args, stdout=self.stdout.fd, stderr=self.stderr.fd, env=env)
        cleanup_funcs.append(lambda: self.kill())

    def finish(self, timeout: float):
        try:
            self.subprocess.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            self.kill()
            raise TestError(self.format_output() + '\n\n' + self.name + ' timed out')

    def kill(self):
        if self.subprocess.returncode is None:
            self.subprocess.kill()
            self.subprocess.wait()

    def format_output(self):
        assert self.subprocess.returncode is not None, 'Program.format_output() called before process exited'
        return format_process_report(
            self.name,
            self.subprocess.returncode,
            self.stdout.collect_str(),
            self.stderr.collect_str())

    def check_returncode(self):
        assert self.subprocess.returncode is not None, repr(self.name) + '.check_returncode() called before process exited'
        if self.subprocess.returncode != 0:
            raise TestError(
                self.format_output() + '\n\n' +
                self.name + ' failed (return code ' + str(self.subprocess.returncode) + ')')

    def collect_output(self):
        return self.stdout.collect_str(), self.stderr.collect_str()

def run_test(name: str, server_args: List[str], client_args: List[str], xdg_runtime: str, wayland_display: str) -> str:
    '''
    Runs two processes: a mock server and the test client
    Does *not* check that client's message assertions pass, this must be done later using the returned output
    '''
    env = os.environ.copy()
    env['XDG_RUNTIME_DIR'] = xdg_runtime
    env['WAYLAND_DISPLAY'] = wayland_display
    env['WAYLAND_DEBUG'] = '1'

    server = Program('server', server_args, env)

    try:
        wait_until_appears(path.join(xdg_runtime, wayland_display))
    except TestError as e:
        server.kill()
        raise TestError(server.format_output() + '\n\n' + str(e))

    client = Program(name, client_args, env)
    client.finish(timeout=10)
    server.finish(timeout=1)

    server.check_returncode()
    client.check_returncode()

    client_stdout, client_stderr = client.collect_output()

    if client_stdout.strip() != '':
        raise TestError(format_stream(name + ' stdout', client_stdout) + '\n\n' + name + ' stdout not empty')

    return client_stderr

def line_contains(line: str, tokens: List[str]) -> bool:
    '''Returns if the given line contains a list of tokens in the given order (anything can be between tokens)'''
    found = True
    for token in tokens:
        if token in line:
            line = line[line.find(token) + len(token):]
        else:
            found = False
    return found

def verify_result(lines: List[str]):
    '''Runs through the output of a client and verifies that all expectations pass, see the test README.md details'''
    assertions = []
    section_start = 0
    for i, line in enumerate(lines):
        if line.startswith('EXPECT: '):
            assertions.append(line.split()[1:])
        elif line.startswith('[') and line.endswith(')') and '@' in line:
            if assertions and line_contains(line, assertions[0]):
                assertions = assertions[1:]
        elif line == 'CHECK EXPECTATIONS COMPLETED' or i == len(lines) - 1:
            if assertions:
                section = format_stream('relevant section', '\n'.join(lines[section_start:i]))
                raise TestError(section + '\n\ndid not find "' + ' '.join(assertions[0]) + '"')
            section_start = i + 1

def main():
    client_bin = sys.argv[1]
    name = path.basename(client_bin)
    server_bin = path.join(path.dirname(client_bin), 'mock-server', 'mock-server')
    assert path.exists(client_bin), 'Could not find client at ' + client_bin
    assert path.exists(server_bin), 'Could not find server at ' + server_bin
    wayland_display = 'wayland-test'
    xdg_runtime = get_xdg_runtime_dir()

    client_stderr = run_test(name, [server_bin], [client_bin, '--auto'], xdg_runtime, wayland_display)
    client_lines = [line.strip() for line in client_stderr.strip().splitlines()]

    try:
        verify_result(client_lines)
    except TestError as e:
        raise TestError(format_stream(name + ' stderr', client_stderr) + '\n\n' + str(e))

if __name__ == '__main__':
    assert len(sys.argv) == 2, 'Incorrect number of args. ' + usage
    fail = False
    try:
        main()
        print('Passed')
    except TestError as e:
        fail = True
        print(e)
    finally:
        for func in cleanup_funcs:
            func()
    if fail:
        exit(1)
